#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import dataclasses
import typing

import sipmessage

from .transaction import TransactionLayer
from .transport import RequestHandler
from .utils import (
    create_contact,
    create_response,
    create_via,
    random_string,
    response_establishes_dialog,
)

DialogKey = tuple[str, str]

# When we drop support for Python 3.10, we can use `typing.Self`.
T = typing.TypeVar("T", bound="Dialog")


def get_dialog_key(request: sipmessage.Request) -> DialogKey:
    call_id = request.call_id or ""
    tag = request.to_address.parameters.get("tag") or ""
    return (call_id, tag)


def replace_tag(address: sipmessage.Address, tag: str) -> sipmessage.Address:
    return dataclasses.replace(
        address,
        parameters=address.parameters.replace(tag=tag),
    )


class Dialog:
    """
    A SIP dialog, representing a peer-to-peer SIP relationship between two
    user agents that persists for some time.

    See :rfc:`3261#section-12`.
    """

    def __init__(
        self,
        *,
        call_id: str,
        dialog_layer: "DialogLayer",
        local_address: sipmessage.Address,
        remote_address: sipmessage.Address,
        route_set: list[sipmessage.Address] = [],
    ) -> None:
        self.call_id = call_id
        self._dialog_layer = dialog_layer
        # Set the tag for the local address unless we were given one.
        local_tag = local_address.parameters.get("tag")
        if local_tag is None:
            local_tag = random_string(10)
            self.local_address = replace_tag(local_address, local_tag)
        self.local_cseq = 1
        self.remote_address = remote_address
        self.route_set = route_set

        # Register the dialog.
        dialog_layer._dialogs[(self.call_id, local_tag)] = self

    @classmethod
    def create_uac(
        cls: typing.Type[T],
        *,
        dialog_layer: "DialogLayer",
        local_address: sipmessage.Address,
        remote_address: sipmessage.Address,
        route_set: list[sipmessage.Address] = [],
        **kwargs: object,
    ) -> T:
        """
        Create a client dialog.

        The initial request will neeed to be constructed using :meth:`create_request`.
        """
        return cls(
            call_id=random_string(22),
            dialog_layer=dialog_layer,
            local_address=local_address,
            remote_address=remote_address,
            route_set=route_set,
            **kwargs,
        )

    @classmethod
    def create_uas(
        cls: typing.Type[T],
        *,
        dialog_layer: "DialogLayer",
        request: sipmessage.Request,
        **kwargs: object,
    ) -> T:
        """
        Create a server dialog based on a received request.
        """
        return cls(
            call_id=request.call_id,
            dialog_layer=dialog_layer,
            local_address=request.to_address,
            remote_address=request.from_address,
            # Populate the route set from the Record-Route headers.
            route_set=request.record_route,
            **kwargs,
        )

    def create_request(self, method: str) -> sipmessage.Request:
        """
        Create a new request for this dialog.
        """
        request = sipmessage.Request(method, self.remote_address.uri)
        request.via = [create_via()]
        request.max_forwards = 70
        request.to_address = self.remote_address
        request.from_address = self.local_address
        request.call_id = self.call_id
        request.cseq = sipmessage.CSeq(sequence=self.local_cseq, method=method)
        request.contact = [create_contact()]
        request.route = self.route_set
        self.local_cseq += 1
        return request

    def create_response(
        self, request: sipmessage.Request, code: int
    ) -> sipmessage.Response:
        """
        Create a response to the given request.
        """
        response = create_response(request=request, code=code)
        if response_establishes_dialog(request.method, code):
            response.to_address = self.local_address
            response.contact = [create_contact()]
        return response

    async def handle_request(self, request: sipmessage.Request) -> None:
        """
        Handle a request for this dialog.

        This needs to be overriden as the base implementation replies
        with a "Not Implemented" error.
        """
        response = self.create_response(request, 501)
        await self.send_response(response)

    async def send_request(self, request: sipmessage.Request) -> sipmessage.Response:
        """
        Send the request in a transaction and return the final response.
        """
        response = await self._dialog_layer._transaction_layer.send_request(request)

        # See https://datatracker.ietf.org/doc/html/rfc3261#section-12.1.2
        if response_establishes_dialog(response.cseq.method, response.code):
            # Set the tag for the remote address.
            tag = response.to_address.parameters["tag"]
            assert isinstance(tag, str)
            self.remote_address = replace_tag(self.remote_address, tag)

            # Populate the route set from the Record-Route headers taken
            # in reverse order.
            self.route_set = list(reversed(response.record_route))

        return response

    async def send_response(self, response: sipmessage.Response) -> None:
        """
        Send the response.
        """
        await self._dialog_layer._transaction_layer.send_response(response)


class DialogLayer:
    """
    SIP transaction layer.

    See :rfc:`3261#section-12`.
    """

    def __init__(self, transaction_layer: TransactionLayer) -> None:
        self._dialogs: dict[DialogKey, Dialog] = {}
        self._transaction_layer = transaction_layer
        self._transaction_layer.request_handler = self._handle_request

        #: A coroutine which will be called whenever a request is received
        #: which does not match an existing dialog.
        self.request_handler: RequestHandler | None = None

    async def _handle_request(self, request: sipmessage.Request) -> None:
        # Deliver the request to the dialog or the application.
        key = get_dialog_key(request)
        dialog = self._dialogs.get(key)
        if dialog is not None:
            await dialog.handle_request(request)
        elif self.request_handler is not None:
            await self.request_handler(request)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "sipua"
copyright = "2025, Spacinov Engineering"
author = "Spacinov Engineering"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
]
intersphinx_mapping = {
    "aiortc": ("https://aiortc.readthedocs.org/en/stable", None),
    "python": ("https://docs.python.org/3", None),
    "sipmessage": ("https://sipmessage.readthedocs.org/en/stable", None),
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import asyncio
import contextlib
import dataclasses
import socket
import unittest
from collections.abc import AsyncGenerator
from unittest.mock import patch

import sipmessage
import websockets.asyncio.client
import websockets.asyncio.connection
from sipua.transport import (
    WEBSOCKET_SUBPROTOCOL,
    TransportAddress,
    TransportLayer,
    get_transport_destination,
    serialize_message,
    update_request_via,
)
from sipua.utils import create_contact, create_response, create_via

from .utils import asynctest, lf2crlf, parse_response


def create_request(
    *,
    uri_port: int,
    uri_transport: str | None = None,
) -> sipmessage.Request:
    uri = sipmessage.URI(scheme="sip", user="bob", host="127.0.0.1", port=uri_port)
    if uri_transport is not None:
        uri = dataclasses.replace(
            uri, parameters=sipmessage.Parameters(transport=uri_transport)
        )
    request = sipmessage.Request(method="OPTIONS", uri=uri)
    with patch("sipua.utils.random_string", new=lambda x: "1e5b2b763d"):
        request.via = [create_via()]
    request.max_forwards = 70
    request.to_address = sipmessage.Address.parse("sip:bob@example.com")
    request.from_address = sipmessage.Address.parse(
        "sip:alice@example.com;tag=7bc759c98ae3e112"
    )
    request.call_id = "126a8db08eba7fb6"
    request.cseq = sipmessage.CSeq(1, "OPTIONS")
    request.contact = [create_contact()]
    return request


def create_request_bytes(
    *,
    via_addr: str = "1.2.3.4:12345",
    via_transport: str,
) -> bytes:
    return lf2crlf(
        f"""INVITE sip:bob@example.com SIP/2.0
Via: SIP/2.0/{via_transport} {via_addr};branch=z9hG4bK1e5b2b763d;rport
Max-Forwards: 70
To: sip:bob@example.com
From: sip:alice@example.com;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 OPTIONS
Content-Length: 0

    """.encode()
    )


def create_response_bytes(*, via_port: int, via_transport: str) -> bytes:
    return lf2crlf(
        f"""SIP/2.0 200 OK
Via: SIP/2.0/{via_transport} 127.0.0.1:{via_port};branch=z9hG4bK1e5b2b763d
To: <sip:bob@example.com>
From: <sip:alice@example.com>;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 OPTIONS
Content-Length: 0

""".encode()
    )


class TestSocket:
    local_port: int
    remote_port: int

    def __init__(self, sock: socket.socket) -> None:
        self.local_port = sock.getsockname()[1]
        self.remote_port = sock.getpeername()[1]
        self._sock = sock

    async def close(self) -> None:
        self._sock.close()

    async def recv(self) -> bytes:
        return self._sock.recv(2000)

    async def send(self, data: bytes) -> None:
        self._sock.send(data)
        await asyncio.sleep(0.1)


class TestWebsocket:
    local_port: int

    def __init__(self, sock: websockets.asyncio.connection.Connection) -> None:
        self.local_port = sock.local_address[1]
        self.remote_port = sock.remote_address[1]
        self._sock = sock

    async def close(self) -> None:
        await self._sock.close()

    async def recv(self) -> bytes:
        data = await self._sock.recv(decode=False)
        assert isinstance(data, bytes)
        return data

    async def send(self, data: bytes) -> None:
        await self._sock.send(data)
        await asyncio.sleep(0.1)


class BaseTestCase(unittest.TestCase):
    def assertMessage(self, message: sipmessage.Message, data: bytes) -> None:
        self.assertEqual(serialize_message(message), lf2crlf(data))

    @contextlib.asynccontextmanager
    async def transport_layer(
        self,
        listen_addresses: list[TransportAddress],
    ) -> AsyncGenerator[tuple[TransportLayer, list[sipmessage.Message]], None]:
        received: list[sipmessage.Message] = []

        async def message_handler(message: sipmessage.Message) -> None:
            received.append(message)

        transport = TransportLayer()
        transport.request_handler = message_handler
        transport.response_handler = message_handler
        try:
            for addr in listen_addresses:
                await transport.listen(addr)
            yield transport, received
        finally:
            await transport.close()


class NoTransportChannelTest(BaseTestCase):
    @asynctest
    async def test_handle_response_no_vias(self) -> None:
        response = parse_response(b"""SIP/2.0 200 OK
To: sip:+33233445566@127.0.0.1:5060
From: sip:+33122334455@127.0.0.1:43248;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 INVITE

""")
        async with self.transport_layer([]) as (transport, _received):
            with self.assertLogs() as cm:
                await transport._handle_message(response)
        self.assertEqual(
            cm.output,
            ["WARNING:sipua.transport:Expected exactly one Via header, got 0"],
        )

    @asynctest
    async def test_handle_response_too_many_vias(self) -> None:
        response = parse_response(b"""SIP/2.0 200 OK
Via: SIP/2.0/UDP 127.0.0.1:43248;branch=z9hG4bK1e5b2b763d
Via: SIP/2.0/UDP 127.0.0.1:1234;branch=z9hG4bKabcdefghij
To: sip:+33233445566@127.0.0.1:5060
From: sip:+33122334455@127.0.0.1:43248;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 INVITE

""")
        async with self.transport_layer([]) as (transport, _received):
            with self.assertLogs() as cm:
                await transport._handle_message(response)
        self.assertEqual(
            cm.output,
            ["WARNING:sipua.transport:Expected exactly one Via header, got 2"],
        )

    @asynctest
    async def test_send_request(self) -> None:
        async with self.transport_layer([]) as (transport, _received):
            request = create_request(uri_port=1234)
            with self.assertRaises(ConnectionError) as cm:
                await transport.send_message(request)
            self.assertEqual(str(cm.exception), "No suitable transport found")


class TcpTransportChannelTest(BaseTestCase):
    @contextlib.asynccontextmanager
    async def transport_layer_and_socket(
        self,
    ) -> AsyncGenerator[
        tuple[TransportLayer, TestSocket, list[sipmessage.Message]], None
    ]:
        async with self.transport_layer(
            [TransportAddress(protocol="tcp", host="127.0.0.1", port=5060)]
        ) as (transport, received):
            with socket.socket(type=socket.SOCK_STREAM) as sock:
                sock.bind(("127.0.0.1", 0))
                sock.connect(("127.0.0.1", 5060))
                yield (transport, TestSocket(sock), received)

    @asynctest
    async def test_receive_garbage(self) -> None:
        async with self.transport_layer_and_socket() as (transport, sock, received):
            await sock.send(b"garbage")

            # Check no message was received.
            self.assertEqual(received, [])

            # Check the channel was closed.
            self.assertEqual(transport._channels, set())

    @asynctest
    async def test_receive_request_send_response(self) -> None:
        async with self.transport_layer_and_socket() as (transport, sock, received):
            # Receive request.
            await sock.send(create_request_bytes(via_transport="TCP"))

            self.assertEqual(len(received), 1)
            request = received[0]
            assert isinstance(request, sipmessage.Request)
            self.assertEqual(request.method, "INVITE")
            self.assertEqual(
                request.uri,
                sipmessage.URI(
                    scheme="sip",
                    host="example.com",
                    user="bob",
                ),
            )
            self.assertEqual(
                request.via,
                [
                    sipmessage.Via(
                        transport="TCP",
                        host="1.2.3.4",
                        port=12345,
                        parameters=sipmessage.Parameters(
                            branch="z9hG4bK1e5b2b763d",
                            rport=str(sock.local_port),
                            received="127.0.0.1",
                        ),
                    )
                ],
            )

            # Send response.
            response = create_response(request=request, code=200)
            is_reliable = await transport.send_message(response)
            self.assertTrue(is_reliable)

            data = await sock.recv()
            self.assertEqual(
                data,
                lf2crlf(
                    f"""SIP/2.0 200 OK
Via: SIP/2.0/TCP 1.2.3.4:12345;branch=z9hG4bK1e5b2b763d;rport={sock.local_port};received=127.0.0.1
To: <sip:bob@example.com>
From: <sip:alice@example.com>;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 OPTIONS
Content-Length: 0

""".encode()
                ),
            )

    @asynctest
    async def test_send_request_and_receive_response(self) -> None:
        async with self.transport_layer(
            [TransportAddress(protocol="tcp", host="127.0.0.1", port=5060)]
        ) as (transport, received):
            with socket.socket(type=socket.SOCK_STREAM) as server_sock:
                server_sock.bind(("127.0.0.1", 0))
                server_sock.listen()

                # Send request.
                request = create_request(
                    uri_port=server_sock.getsockname()[1],
                    uri_transport="tcp",
                )
                await transport.send_message(request)

                sock = TestSocket(server_sock.accept()[0])
                data = await sock.recv()
                self.assertEqual(
                    data,
                    lf2crlf(
                        f"""OPTIONS sip:bob@127.0.0.1:{sock.local_port};transport=tcp SIP/2.0
Via: SIP/2.0/TCP 127.0.0.1:{sock.remote_port};branch=z9hG4bK1e5b2b763d
Max-Forwards: 70
To: <sip:bob@example.com>
From: <sip:alice@example.com>;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 OPTIONS
Contact: <sip:127.0.0.1:{sock.remote_port};transport=tcp>
Content-Length: 0

""".encode()
                    ),
                )

                # Receive response.
                await sock.send(
                    create_response_bytes(
                        via_port=sock.remote_port, via_transport="TCP"
                    )
                )

                self.assertEqual(len(received), 1)
                response = received[0]
                assert isinstance(response, sipmessage.Response)
                self.assertEqual(response.code, 200)
                self.assertEqual(response.phrase, "OK")

                await sock.close()

    @asynctest
    async def test_connect_fails(self) -> None:
        async with (
            self.transport_layer(
                [TransportAddress(protocol="tcp", host="127.0.0.1", port=5060)]
            ) as (transport, received),
        ):
            # Client sends request.
            request = create_request(uri_port=5080, uri_transport="tcp")
            with self.assertRaises(ConnectionRefusedError):
                await transport.send_message(request)


class UdpTransportChannelTest(BaseTestCase):
    @contextlib.asynccontextmanager
    async def transport_layer_and_socket(
        self,
    ) -> AsyncGenerator[
        tuple[TransportLayer, TestSocket, list[sipmessage.Message]], None
    ]:
        async with self.transport_layer(
            [TransportAddress(protocol="udp", host="127.0.0.1", port=5060)]
        ) as (transport, received):
            with socket.socket(type=socket.SOCK_DGRAM) as sock:
                sock.bind(("127.0.0.1", 0))
                sock.connect(("127.0.0.1", 5060))
                yield (transport, TestSocket(sock), received)

    @asynctest
    async def test_receive_garbage(self) -> None:
        async with self.transport_layer_and_socket() as (transport, sock, received):
            await sock.send(b"garbage")

            # Check no message was received.
            self.assertEqual(received, [])

            # Check the channel was not closed.
            self.assertEqual(len(transport._channels), 1)

    @asynctest
    async def test_receive_request_send_response(self) -> None:
        async with self.transport_layer_and_socket() as (transport, sock, received):
            # Receive request.
            await sock.send(create_request_bytes(via_transport="UDP"))

            self.assertEqual(len(received), 1)
            request = received[0]
            assert isinstance(request, sipmessage.Request)
            self.assertEqual(request.method, "INVITE")
            self.assertEqual(
                request.uri,
                sipmessage.URI(
                    scheme="sip",
                    host="example.com",
                    user="bob",
                ),
            )
            self.assertEqual(
                request.via,
                [
                    sipmessage.Via(
                        transport="UDP",
                        host="1.2.3.4",
                        port=12345,
                        parameters=sipmessage.Parameters(
                            branch="z9hG4bK1e5b2b763d",
                            rport=str(sock.local_port),
                            received="127.0.0.1",
                        ),
                    )
                ],
            )

            # Send response.
            response = create_response(request=request, code=200)
            is_reliable = await transport.send_message(response)
            self.assertFalse(is_reliable)

            data = await sock.recv()
            self.assertEqual(
                data,
                lf2crlf(
                    f"""SIP/2.0 200 OK
Via: SIP/2.0/UDP 1.2.3.4:12345;branch=z9hG4bK1e5b2b763d;rport={sock.local_port};received=127.0.0.1
To: <sip:bob@example.com>
From: <sip:alice@example.com>;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 OPTIONS
Content-Length: 0

""".encode()
                ),
            )

    @asynctest
    async def test_send_request_and_receive_response(self) -> None:
        async with self.transport_layer_and_socket() as (transport, sock, received):
            # Send request.
            request = create_request(uri_port=sock.local_port)
            await transport.send_message(request)

            data = await sock.recv()
            self.assertEqual(
                data,
                lf2crlf(
                    f"""OPTIONS sip:bob@127.0.0.1:{sock.local_port} SIP/2.0
Via: SIP/2.0/UDP 127.0.0.1:5060;branch=z9hG4bK1e5b2b763d
Max-Forwards: 70
To: <sip:bob@example.com>
From: <sip:alice@example.com>;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 OPTIONS
Contact: <sip:127.0.0.1:5060;transport=udp>
Content-Length: 0

""".encode()
                ),
            )

            # Receive response.
            await sock.send(create_response_bytes(via_port=5060, via_transport="UDP"))

            self.assertEqual(len(received), 1)
            response = received[0]
            assert isinstance(response, sipmessage.Response)
            self.assertEqual(response.code, 200)
            self.assertEqual(response.phrase, "OK")


class WebsocketTransportChannelTest(BaseTestCase):
    @contextlib.asynccontextmanager
    async def transport_layer_and_socket(
        self,
    ) -> AsyncGenerator[
        tuple[
            TransportLayer,
            TestWebsocket,
            list[sipmessage.Message],
        ],
        None,
    ]:
        async with self.transport_layer(
            [TransportAddress(protocol="ws", host="127.0.0.1", port=5080)]
        ) as (transport, received):
            async with websockets.asyncio.client.connect(
                "ws://127.0.0.1:5080", subprotocols=[WEBSOCKET_SUBPROTOCOL]
            ) as websocket:
                yield (transport, TestWebsocket(websocket), received)

    @asynctest
    async def test_receive_garbage(self) -> None:
        async with self.transport_layer_and_socket() as (transport, sock, received):
            await sock.send(b"garbage")

            # Check no message was received.
            self.assertEqual(received, [])

            # Check the connection was not closed.
            self.assertEqual(len(transport._channels), 1)

    @asynctest
    async def test_receive_request_send_response(self) -> None:
        async with self.transport_layer_and_socket() as (transport, sock, received):
            # Receive request.
            await sock.send(
                create_request_bytes(
                    via_addr="li62vs2t75f6.invalid", via_transport="WS"
                )
            )

            self.assertEqual(len(received), 1)
            request = received[0]
            assert isinstance(request, sipmessage.Request)
            self.assertEqual(request.method, "INVITE")
            self.assertEqual(
                request.uri,
                sipmessage.URI(
                    scheme="sip",
                    host="example.com",
                    user="bob",
                ),
            )
            self.assertEqual(
                request.via,
                [
                    sipmessage.Via(
                        transport="WS",
                        host="li62vs2t75f6.invalid",
                        parameters=sipmessage.Parameters(
                            branch="z9hG4bK1e5b2b763d",
                            rport=str(sock.local_port),
                            received="127.0.0.1",
                        ),
                    )
                ],
            )

            # Send response.
            response = create_response(request=request, code=200)
            is_reliable = await transport.send_message(response)
            self.assertTrue(is_reliable)

            data = await sock.recv()
            self.assertEqual(
                data,
                lf2crlf(
                    f"""SIP/2.0 200 OK
Via: SIP/2.0/WS li62vs2t75f6.invalid;branch=z9hG4bK1e5b2b763d;rport={sock.local_port};received=127.0.0.1
To: <sip:bob@example.com>
From: <sip:alice@example.com>;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 OPTIONS
Content-Length: 0

""".encode()
                ),
            )

    @asynctest
    @patch("sipua.transport.random_string", new=lambda x: "WeNzcJ9a6ATr")  # type: ignore
    async def test_send_request_and_receive_response(self) -> None:
        async with self.transport_layer_and_socket() as (transport, sock, received):
            # Send request.
            request = create_request(uri_port=sock.local_port, uri_transport="ws")
            await transport.send_message(request)

            data = await sock.recv()
            self.assertEqual(
                data,
                lf2crlf(
                    f"""OPTIONS sip:bob@127.0.0.1:{sock.local_port};transport=ws SIP/2.0
Via: SIP/2.0/WS WeNzcJ9a6ATr.invalid;branch=z9hG4bK1e5b2b763d
Max-Forwards: 70
To: <sip:bob@example.com>
From: <sip:alice@example.com>;tag=7bc759c98ae3e112
Call-ID: 126a8db08eba7fb6
CSeq: 1 OPTIONS
Contact: <sip:WeNzcJ9a6ATr.invalid;transport=ws>
Content-Length: 0

""".encode()
                ),
            )

            # Receive response.
            await sock.send(create_response_bytes(via_port=5060, via_transport="UDP"))

            self.assertEqual(len(received), 1)
            response = received[0]
            assert isinstance(response, sipmessage.Response)
            self.assertEqual(response.code, 200)
            self.assertEqual(response.phrase, "OK")

    @asynctest
    async def test_connect_websocket(self) -> None:
        async with (
            self.transport_layer(
                [TransportAddress(protocol="ws", host="127.0.0.1", port=5080)]
            ) as (server_transport, server_received),
            self.transport_layer([]) as (client_transport, client_received),
        ):
            await client_transport.connect_websocket("ws://127.0.0.1:5080")

            # Client sends request.
            request = create_request(uri_port=5080, uri_transport="ws")
            await client_transport.send_message(request)
            await asyncio.sleep(0.1)

            self.assertEqual(len(server_received), 1)
            received_request = server_received[0]
            assert isinstance(received_request, sipmessage.Request)

            # Server sends reply.
            response = create_response(request=received_request, code=200)
            await server_transport.send_message(response)
            await asyncio.sleep(0.1)

            self.assertEqual(len(client_received), 1)
            client_response = client_received[0]
            assert isinstance(client_response, sipmessage.Response)


class GetTransportDestinationTest(unittest.TestCase):
    @asynctest
    async def test_request_sip(self) -> None:
        request = sipmessage.Request(
            method="INVITE", uri=sipmessage.URI(scheme="sip", host="1.2.3.4")
        )
        self.assertEqual(
            await get_transport_destination(request),
            TransportAddress(protocol="udp", host="1.2.3.4", port=5060),
        )

    @asynctest
    async def test_request_sip_loose_routing(self) -> None:
        """
        A Route is specified, using loose routing.
        """
        request = sipmessage.Request(
            method="INVITE", uri=sipmessage.URI(scheme="sip", host="1.2.3.4")
        )
        request.route = [
            sipmessage.Address(
                uri=sipmessage.URI(
                    scheme="sip",
                    host="2.3.4.5",
                    parameters=sipmessage.Parameters(lr=None),
                )
            )
        ]
        self.assertEqual(
            await get_transport_destination(request),
            TransportAddress(protocol="udp", host="2.3.4.5", port=5060),
        )

    @asynctest
    async def test_request_sip_strict_routing(self) -> None:
        """
        A Route is specified, using strict routing.
        """
        request = sipmessage.Request(
            method="INVITE", uri=sipmessage.URI(scheme="sip", host="1.2.3.4")
        )
        request.route = [
            sipmessage.Address(uri=sipmessage.URI(scheme="sip", host="2.3.4.5"))
        ]
        self.assertEqual(
            await get_transport_destination(request),
            TransportAddress(protocol="udp", host="1.2.3.4", port=5060),
        )

    @asynctest
    async def test_request_sip_transport_udp(self) -> None:
        """
        UDP transport is explicitly specified in the Request-URI and
        the host is a numeric IP address.
        """
        request = sipmessage.Request(
            method="INVITE",
            uri=sipmessage.URI(
                scheme="sip",
                host="1.2.3.4",
                parameters=sipmessage.Parameters(transport="udp"),
            ),
        )
        self.assertEqual(
            await get_transport_destination(request),
            TransportAddress(protocol="udp", host="1.2.3.4", port=5060),
        )

    @asynctest
    async def test_request_sip_transport_udp_fqdn(self) -> None:
        """
        UDP transport is explicitly specified in the Request-URI and
        the host is an FQDN.
        """
        request = sipmessage.Request(
            method="INVITE",
            uri=sipmessage.URI(
                scheme="sip",
                host="atlanta.com",
                parameters=sipmessage.Parameters(transport="udp"),
            ),
        )
        with self.assertRaises(NotImplementedError) as cm:
            await get_transport_destination(request)
        self.assertEqual(str(cm.exception), "SRV lookups are not yet supported")

    @asynctest
    async def test_request_sip_transport_udp_fqdn_with_port(self) -> None:
        """
        UDP transport is explicitly specified in the Request-URI and
        the host is an FQDN.
        """
        request = sipmessage.Request(
            method="INVITE",
            uri=sipmessage.URI(
                scheme="sip",
                host="atlanta.com",
                port=5060,
                parameters=sipmessage.Parameters(transport="udp"),
            ),
        )
        with patch("socket.gethostbyname") as mock_gethostbyname:
            mock_gethostbyname.return_value = "1.2.3.4"
            self.assertEqual(
                await get_transport_destination(request),
                TransportAddress(protocol="udp", host="1.2.3.4", port=5060),
            )

    @asynctest
    async def test_request_sip_transport_tcp(self) -> None:
        """
        TCP transport is explicitly specified in the Request-URI and
        the host is a numeric IP address.
        """
        request = sipmessage.Request(
            method="INVITE",
            uri=sipmessage.URI(
                scheme="sip",
                host="1.2.3.4",
                parameters=sipmessage.Parameters(transport="tcp"),
            ),
        )
        self.assertEqual(
            await get_transport_destination(request),
            TransportAddress(protocol="tcp", host="1.2.3.4", port=5060),
        )

    @asynctest
    async def test_request_sip_transport_ws_fqdn(self) -> None:
        """
        WS transport is explicitly specified in the Request-URI and
        the host is an FQDN.
        """
        request = sipmessage.Request(
            method="INVITE",
            uri=sipmessage.URI(
                scheme="sip",
                host="example.com",
                parameters=sipmessage.Parameters(transport="ws"),
            ),
        )
        self.assertEqual(
            await get_transport_destination(request),
            TransportAddress(protocol="ws", host="example.com", port=0),
        )

    @asynctest
    async def test_request_sip_fqdn(self) -> None:
        # NOTE: this should theoretically trigger an NAPTR lookup.
        request = sipmessage.Request(
            method="INVITE", uri=sipmessage.URI(scheme="sip", host="atlanta.com")
        )
        with self.assertRaises(NotImplementedError) as cm:
            await get_transport_destination(request)
        self.assertEqual(str(cm.exception), "SRV lookups are not yet supported")

    @asynctest
    async def test_request_sips(self) -> None:
        request = sipmessage.Request(
            method="INVITE", uri=sipmessage.URI(scheme="sips", host="1.2.3.4")
        )
        self.assertEqual(
            await get_transport_destination(request),
            TransportAddress(protocol="tls", host="1.2.3.4", port=5061),
        )

    @asynctest
    async def test_request_sips_fqdn(self) -> None:
        # NOTE: this should theoretically trigger an NAPTR lookup.
        request = sipmessage.Request(
            method="INVITE", uri=sipmessage.URI(scheme="sips", host="atlanta.com")
        )
        with self.assertRaises(NotImplementedError) as cm:
            await get_transport_destination(request)
        self.assertEqual(str(cm.exception), "SRV lookups are not yet supported")

    @asynctest
    async def test_response_tcp(self) -> None:
        response = sipmessage.Response(code=200, phrase="OK")
        response.via = [sipmessage.Via(transport="TCP", host="1.2.3.4")]
        self.assertEqual(
            await get_transport_destination(response),
            TransportAddress(protocol="tcp", host="1.2.3.4", port=5060),
        )

    @asynctest
    async def test_response_tls(self) -> None:
        response = sipmessage.Response(code=200, phrase="OK")
        response.via = [sipmessage.Via(transport="TLS", host="1.2.3.4")]
        self.assertEqual(
            await get_transport_destination(response),
            TransportAddress(protocol="tls", host="1.2.3.4", port=5061),
        )

    @asynctest
    async def test_response_udp(self) -> None:
        response = sipmessage.Response(code=200, phrase="OK")
        response.via = [sipmessage.Via(transport="UDP", host="1.2.3.4")]
        self.assertEqual(
            await get_transport_destination(response),
            TransportAddress(protocol="udp", host="1.2.3.4", port=5060),
        )

    @asynctest
    async def test_response_udp_with_received_and_rport(self) -> None:
        response = sipmessage.Response(code=200, phrase="OK")
        response.via = [
            sipmessage.Via(
                transport="UDP",
                host="example.com",
                parameters=sipmessage.Parameters(received="1.2.3.4", rport="1234"),
            )
        ]
        self.assertEqual(
            await get_transport_destination(response),
            TransportAddress(protocol="udp", host="1.2.3.4", port=1234),
        )

    @asynctest
    async def test_response_ws(self) -> None:
        response = sipmessage.Response(code=200, phrase="OK")
        response.via = [sipmessage.Via(transport="WS", host="mYn6S3lQaKjo.invalid")]
        self.assertEqual(
            await get_transport_destination(response),
            TransportAddress(protocol="ws", host="mYn6S3lQaKjo.invalid", port=0),
        )

    @asynctest
    async def test_response_wss(self) -> None:
        response = sipmessage.Response(code=200, phrase="OK")
        response.via = [sipmessage.Via(transport="WSS", host="mYn6S3lQaKjo.invalid")]
        self.assertEqual(
            await get_transport_destination(response),
            TransportAddress(protocol="wss", host="mYn6S3lQaKjo.invalid", port=0),
        )


class UpdateRequestViaTest(unittest.TestCase):
    def test_no_via(self) -> None:
        request = sipmessage.Request(
            "INVITE", sipmessage.URI(scheme="sip", user="bob", host="biloxi.com")
        )
        update_request_via(request, "1.2.3.4", 5060)
        self.assertEqual(request.via, [])

    def test_via_received_different_host(self) -> None:
        request = sipmessage.Request(
            "INVITE", sipmessage.URI(scheme="sip", user="bob", host="biloxi.com")
        )
        request.via = [
            sipmessage.Via(
                transport="UDP",
                host="bobspc.biloxi.com",
            )
        ]
        update_request_via(request, "1.2.3.4", 5060)
        self.assertEqual(
            request.via,
            [
                sipmessage.Via(
                    transport="UDP",
                    host="bobspc.biloxi.com",
                    parameters=sipmessage.Parameters(received="1.2.3.4"),
                )
            ],
        )

    def test_via_received_same_host(self) -> None:
        request = sipmessage.Request(
            "INVITE", sipmessage.URI(scheme="sip", user="bob", host="biloxi.com")
        )
        request.via = [sipmessage.Via(transport="UDP", host="1.2.3.4")]
        update_request_via(request, "1.2.3.4", 5060)
        self.assertEqual(request.via, [sipmessage.Via(transport="UDP", host="1.2.3.4")])

    def test_via_rport(self) -> None:
        request = sipmessage.Request(
            "INVITE", sipmessage.URI(scheme="sip", user="bob", host="biloxi.com")
        )
        request.via = [
            sipmessage.Via(
                transport="UDP",
                host="1.2.3.4",
                parameters=sipmessage.Parameters(rport=None),
            )
        ]
        update_request_via(request, "1.2.3.4", 5060)
        self.assertEqual(
            request.via,
            [
                sipmessage.Via(
                    transport="UDP",
                    host="1.2.3.4",
                    parameters=sipmessage.Parameters(rport="5060"),
                )
            ],
        )



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#


import contextlib
import functools
import unittest
from collections.abc import AsyncGenerator

import sipmessage
import sipua

from .utils import asynctest

ALICE = sipmessage.Address(
    name="Alice",
    uri=sipmessage.URI(scheme="sip", user="alice", host="127.0.0.1", port=5061),
)
BOB = sipmessage.Address(
    name="Bob",
    uri=sipmessage.URI(scheme="sip", user="bob", host="127.0.0.1", port=5062),
)


class DialogTest(unittest.TestCase):
    @contextlib.asynccontextmanager
    async def client_and_server(
        self,
    ) -> AsyncGenerator[tuple[sipua.DialogLayer, sipua.DialogLayer], None]:
        client_transport = sipua.TransportLayer()
        client_transaction = sipua.TransactionLayer(client_transport)
        client_dialog = sipua.DialogLayer(client_transaction)

        server_transport = sipua.TransportLayer()
        server_transaction = sipua.TransactionLayer(server_transport)
        server_dialog = sipua.DialogLayer(server_transaction)

        try:
            await client_transport.listen(
                sipua.TransportAddress("udp", "127.0.0.1", 5061)
            )
            await server_transport.listen(
                sipua.TransportAddress("udp", "127.0.0.1", 5062)
            )
            yield (client_dialog, server_dialog)
        finally:
            await client_transport.close()
            await server_transport.close()

    async def server_request_handler(
        self,
        request: sipmessage.Request,
        *,
        dialog_layer: sipua.DialogLayer,
        route_set: list[sipmessage.Address] = [],
    ) -> None:
        self.assertEqual(request.method, "INVITE")

        # Server creates dialog and responds 200.
        server_dialog = sipua.Dialog.create_uas(
            dialog_layer=dialog_layer, request=request
        )
        response = server_dialog.create_response(request, 200)
        self.assertEqual(response.call_id, request.call_id)
        self.assertEqual(response.cseq, request.cseq)
        self.assertEqual(response.from_address, request.from_address)
        self.assertEqual(response.via, request.via)
        transaction_layer = dialog_layer._transaction_layer
        await transaction_layer.send_response(response)

        # Check server dialog state.
        self.assertEqual(server_dialog.local_cseq, 1)
        self.assertEqual(server_dialog.remote_address, request.from_address)
        self.assertEqual(server_dialog.route_set, route_set)

    @asynctest
    async def test_invite(self) -> None:
        async with self.client_and_server() as (client, server):
            server.request_handler = functools.partial(
                self.server_request_handler, dialog_layer=server
            )

            # Client sends INVITE, server responds 200.
            client_dialog = sipua.Dialog.create_uac(
                dialog_layer=client,
                local_address=ALICE,
                remote_address=BOB,
            )
            request = client_dialog.create_request("INVITE")
            self.assertEqual(request.cseq, sipmessage.CSeq(sequence=1, method="INVITE"))
            response = await client_dialog.send_request(request)
            self.assertEqual(response.code, 200)

            # Check client dialog state.
            self.assertEqual(client_dialog.local_cseq, 2)
            self.assertEqual(client_dialog.remote_address, response.to_address)
            self.assertEqual(client_dialog.route_set, [])

            # Client sends BYE, server responds 501.
            request = client_dialog.create_request("BYE")
            response = await client_dialog.send_request(request)
            self.assertEqual(response.code, 501)



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import importlib.metadata

from .application import Application
from .call import Call
from .dialog import Dialog, DialogLayer
from .transaction import TransactionLayer
from .transport import TransportAddress, TransportLayer
from .utils import create_ack, create_response

__all__ = [
    "create_ack",
    "create_response",
    "Application",
    "Call",
    "Dialog",
    "DialogLayer",
    "TransactionLayer",
    "TransportAddress",
    "TransportLayer",
]
__version__ = importlib.metadata.version("sipua")



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import argparse
import asyncio
import logging

import sipmessage
import sipua
from aiortc.contrib.media import MediaBlackhole, MediaPlayer, MediaRecorder


class SipClient(sipua.Application):
    async def invite(
        self,
        *,
        local_address: sipmessage.Address,
        remote_address: sipmessage.Address,
        play_from: str,
        record_to: str,
    ) -> None:
        call = sipua.Call.create_uac(
            dialog_layer=self.dialog_layer,
            local_address=local_address,
            remote_address=remote_address,
        )

        # Play audio from a file or fall back to silence.
        if play_from:
            player = MediaPlayer(play_from)
            call.audioTransceiver.sender.replaceTrack(player.audio)

        # Consume received audio.
        recorder: MediaBlackhole | MediaRecorder
        if record_to:
            recorder = MediaRecorder(record_to)
        else:
            recorder = MediaBlackhole()
        recorder.addTrack(call.audioTransceiver.receiver.track)
        await recorder.start()

        # Make the call.
        await call.invite()

        # Send media for a few seconds.
        await asyncio.sleep(5)

        # Hangup the call.
        await call.hangup()

        # Stop recording.
        await recorder.stop()


async def main() -> None:
    parser = argparse.ArgumentParser("sip client")
    parser.add_argument("--local-username", type=str, default="+33122334455")
    parser.add_argument("--local-host", type=str, default="127.0.0.1")
    parser.add_argument("--local-port", type=int, default=5061)
    parser.add_argument(
        "--play-from", type=str, help="Read the media from a file and sent it."
    )
    parser.add_argument("--record-to", type=str, help="Write received media to a file.")
    parser.add_argument("--remote-username", type=str, default="+33233445566")
    parser.add_argument("--remote-host", type=str, default="127.0.0.1")
    parser.add_argument("--remote-port", type=int, default=5060)
    options = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    app = SipClient()
    await app.listen(
        sipua.TransportAddress("udp", options.local_host, options.local_port)
    )
    await app.invite(
        local_address=sipmessage.Address(
            uri=sipmessage.URI(
                scheme="sip",
                user=options.local_username,
                host=options.local_host,
                port=options.local_port,
            )
        ),
        remote_address=sipmessage.Address(
            uri=sipmessage.URI(
                scheme="sip",
                user=options.remote_username,
                host=options.remote_host,
                port=options.remote_port,
            )
        ),
        # Media playback and recording.
        play_from=options.play_from,
        record_to=options.record_to,
    )
    await app.close()
    await asyncio.sleep(1)


if __name__ == "__main__":
    asyncio.run(main())



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#


import asyncio
import re
import unittest

import sipmessage
import sipua

from .utils import asynctest

UDP_CLIENT = sipua.TransportAddress("udp", "127.0.0.1", 5061)
UDP_SERVER = sipua.TransportAddress("udp", "127.0.0.1", 5060)


def strip_ice(sdp: str) -> str:
    return re.sub(
        r"^a=(candidate|end-of-candidates|ice-pwd|ice-ufrag).*\n",
        "",
        sdp,
        flags=re.M,
    )


class ClientApplication(sipua.Application):
    async def invite(
        self,
        *,
        local_address: sipmessage.Address,
        remote_address: sipmessage.Address,
        extra_headers: list[tuple[str, str]] = [],
        **kwargs: object,
    ) -> None:
        call = sipua.Call.create_uac(
            dialog_layer=self.dialog_layer,
            local_address=local_address,
            remote_address=remote_address,
            route_set=[],
            **kwargs,
        )
        assert call.audioTransceiver.kind == "audio"

        await call.invite(extra_headers=extra_headers)

        # send media
        await asyncio.sleep(2)

        # hangup
        await call.hangup()

        # hangup again
        await call.hangup()


class ServerApplication(sipua.Application):
    def __init__(self, dtls: bool = False, ice: bool = True) -> None:
        super().__init__()
        self.dtls = dtls
        self.ice = ice
        self.invite_headers: list[str] = []

    async def handle_request(self, request: sipmessage.Request) -> None:
        """
        Handle a request outside a dialog.
        """
        if request.method == "INVITE":
            call = sipua.Call.create_uas(
                dialog_layer=self.dialog_layer,
                dtls=self.dtls,
                request=request,
            )

            # Optionally strip out ICE information from SDP to emulate
            # a peer which does not support ICE.
            if not self.ice:
                real_create_sdp = call._create_sdp
                real_handle_sdp = call._handle_sdp

                async def patched_create_sdp() -> str:
                    sdp = await real_create_sdp()
                    return strip_ice(sdp)

                async def patched_handle_sdp(sdp: str) -> None:
                    sdp = strip_ice(sdp)
                    await real_handle_sdp(sdp)

                call._create_sdp = patched_create_sdp  # type: ignore
                call._handle_sdp = patched_handle_sdp  # type: ignore

            # Send response.
            await call.accept(request)

            self.invite_headers = sorted(request.headers.keys())
        else:
            await super().handle_request(request)


class UdpTest(unittest.TestCase):
    local_address = sipmessage.Address(
        name="Alice",
        uri=sipmessage.URI(
            scheme="sip",
            user="alice",
            host="127.0.0.1",
            port=5061,
        ),
    )
    remote_address = sipmessage.Address(
        name="Bob",
        uri=sipmessage.URI(
            scheme="sip",
            user="bob",
            host="127.0.0.1",
            port=5060,
        ),
    )

    @asynctest
    async def test_invite_ok(self) -> None:
        server = ServerApplication()
        await server.listen(UDP_SERVER)

        client = ClientApplication()
        await client.listen(UDP_CLIENT)
        await client.invite(
            local_address=self.local_address,
            remote_address=self.remote_address,
        )

        await asyncio.gather(client.close(), server.close())

        self.assertEqual(
            server.invite_headers,
            [
                "CSeq",
                "Call-ID",
                "Contact",
                "Content-Length",
                "Content-Type",
                "From",
                "Max-Forwards",
                "To",
                "Via",
            ],
        )

    @asynctest
    async def test_invite_ok_no_ice(self) -> None:
        server = ServerApplication(ice=False)
        await server.listen(UDP_SERVER)

        client = ClientApplication()
        await client.listen(UDP_CLIENT)
        await client.invite(
            local_address=self.local_address,
            remote_address=self.remote_address,
        )

        await asyncio.gather(client.close(), server.close())

    @asynctest
    async def test_invite_ok_with_dtls(self) -> None:
        server = ServerApplication(dtls=True)
        await server.listen(UDP_SERVER)

        client = ClientApplication()
        await client.listen(UDP_CLIENT)
        await client.invite(
            dtls=True,
            local_address=self.local_address,
            remote_address=self.remote_address,
        )

        await asyncio.gather(client.close(), server.close())

    @asynctest
    async def test_invite_ok_with_extra_headers(self) -> None:
        server = ServerApplication(dtls=True)
        await server.listen(UDP_SERVER)

        client = ClientApplication()
        await client.listen(UDP_CLIENT)
        await client.invite(
            dtls=True,
            extra_headers=[
                ("X-Header-A", "value-a"),
                ("X-Header-B", "value-b1"),
                ("X-Header-B", "value-b2"),
            ],
            local_address=self.local_address,
            remote_address=self.remote_address,
        )

        await asyncio.gather(client.close(), server.close())

        self.assertEqual(
            server.invite_headers,
            [
                "CSeq",
                "Call-ID",
                "Contact",
                "Content-Length",
                "Content-Type",
                "From",
                "Max-Forwards",
                "To",
                "Via",
                "X-Header-A",
                "X-Header-B",
            ],
        )

    @asynctest
    async def test_invite_ok_with_stun_server(self) -> None:
        server = ServerApplication()
        await server.listen(UDP_SERVER)

        client = ClientApplication()
        await client.listen(UDP_CLIENT)
        await client.invite(
            local_address=self.local_address,
            remote_address=self.remote_address,
            stun_server="stun:stun.l.google.com:19302",
        )

        await asyncio.gather(client.close(), server.close())

    @asynctest
    async def test_invite_without_handler(self) -> None:
        server = sipua.Application()
        await server.listen(UDP_SERVER)

        client = ClientApplication()
        await client.listen(UDP_CLIENT)
        with self.assertRaises(Exception) as cm:
            await client.invite(
                local_address=self.local_address,
                remote_address=self.remote_address,
            )
        self.assertEqual(str(cm.exception), "Call failed: 501 Not Implemented")

        await asyncio.gather(client.close(), server.close())


class WebSocketTest(unittest.TestCase):
    local_address = sipmessage.Address(
        name="Alice",
        uri=sipmessage.URI(
            scheme="sip",
            user="alice",
            host="example.com",
            parameters=sipmessage.Parameters(transport="ws"),
        ),
    )
    remote_address = sipmessage.Address(
        name="Bob",
        uri=sipmessage.URI(
            scheme="sip",
            user="bob",
            host="example.com",
            parameters=sipmessage.Parameters(transport="ws"),
        ),
    )

    @asynctest
    async def test_invite_ok(self) -> None:
        server = ServerApplication(dtls=False)
        await server.listen(sipua.TransportAddress("ws", "127.0.0.1", 7443))

        client = ClientApplication()
        await client.connect_websocket("ws://127.0.0.1:7443")
        await client.invite(
            local_address=self.local_address,
            remote_address=self.remote_address,
        )

        await asyncio.gather(client.close(), server.close())



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import argparse
import asyncio
import logging

import sipmessage
import sipua
from aiortc.contrib.media import MediaBlackhole, MediaPlayer, MediaRecorder


async def ensure_hangup(
    call: sipua.Call, recorder: MediaBlackhole | MediaRecorder
) -> None:
    await asyncio.sleep(15)
    await call.hangup()
    await recorder.stop()


class SipServer(sipua.Application):
    def __init__(self, play_from: str, record_to: str) -> None:
        super().__init__()
        self.play_from = play_from
        self.record_to = record_to

    async def handle_request(self, request: sipmessage.Request) -> None:
        """
        Handle a request outside a dialog.
        """
        if request.method == "INVITE":
            call = sipua.Call.create_uas(
                dialog_layer=self.dialog_layer,
                request=request,
            )

            # Play audio from a file or fall back to a sine signal.
            if self.play_from:
                player = MediaPlayer(self.play_from)
            else:
                player = MediaPlayer(
                    "sine=frequency=440:sample_rate=48000:duration=1", format="lavfi"
                )
            call.audioTransceiver.sender.replaceTrack(player.audio)

            # Consume received audio.
            recorder: MediaBlackhole | MediaRecorder
            if self.record_to:
                recorder = MediaRecorder(self.record_to)
            else:
                recorder = MediaBlackhole()
            recorder.addTrack(call.audioTransceiver.receiver.track)
            await recorder.start()

            # Accept the incoming call.
            await call.accept(request)

            # Ensure the call is hung up 15s later.
            asyncio.create_task(ensure_hangup(call, recorder))
        else:
            # Reject any other requests.
            await super().handle_request(request)


async def main() -> None:
    parser = argparse.ArgumentParser("sip server")
    parser.add_argument("--local-host", type=str, default="127.0.0.1")
    parser.add_argument("--local-port", type=int, default=5060)
    parser.add_argument(
        "--play-from", type=str, help="Read the media from a file and sent it."
    )
    parser.add_argument("--record-to", type=str, help="Write received media to a file.")
    options = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    server = SipServer(play_from=options.play_from, record_to=options.record_to)
    await server.listen(
        sipua.TransportAddress("udp", options.local_host, options.local_port)
    )
    try:
        await asyncio.Future()
    finally:
        await server.close()


if __name__ == "__main__":
    asyncio.run(main())



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import sipmessage

from .dialog import DialogLayer
from .transaction import TransactionLayer
from .transport import TransportAddress, TransportLayer
from .utils import create_response


class Application:
    """
    Convenience class to facilitate creating a SIP user agent.
    """

    def __init__(self) -> None:
        self.transport_layer = TransportLayer()
        self.transaction_layer = TransactionLayer(self.transport_layer)
        self.dialog_layer = DialogLayer(self.transaction_layer)
        self.dialog_layer.request_handler = self.handle_request

    async def close(self) -> None:
        """
        Close the transport layer.
        """
        await self.transport_layer.close()

    async def connect_websocket(self, addr: str) -> None:
        """
        Establish an outgoing WebSocket connection.
        """
        await self.transport_layer.connect_websocket(addr)

    async def listen(self, address: TransportAddress) -> None:
        """
        Start listening on given transport address.
        """
        await self.transport_layer.listen(address)

    # overrideable

    async def handle_request(self, request: sipmessage.Request) -> None:
        """
        Handle a request which does not match an existing dialog.

        You can override this by subclassing :class:`Application`,
        for instance to handle incoming calls.
        """
        response = create_response(request=request, code=501)
        await self.transaction_layer.send_response(response)



#
# Copyright (C) Spacinov SAS
# Distributed under the 2-clause BSD license
#

import asyncio

import pylibsrtp
import sipmessage
from aiortc import (
    AudioStreamTrack,
    RTCCertificate,
    RTCDtlsParameters,
    RTCDtlsTransport,
    RTCIceCandidate,
    RTCIceGatherer,
    RTCIceParameters,
    RTCIceServer,
    RTCIceTransport,
    RTCRtpCodecParameters,
    RTCRtpReceiver,
    RTCRtpSender,
    RTCRtpTransceiver,
    clock,
    sdp,
)
from aiortc.rtcdtlstransport import State
from aiortc.rtcrtpparameters import RTCRtpReceiveParameters, RTCRtpSendParameters
from aiortc.rtcrtpreceiver import RemoteStreamTrack

from .dialog import Dialog, DialogLayer
from .utils import create_contact

COMPONENT_RTP = 1


class FakeSsl:
    def DTLSv1_get_timeout(self) -> None:
        return None


class FakeDtlsSession(pylibsrtp.Session):
    def protect(self, data: bytes) -> bytes:
        return data

    def protect_rtcp(self, data: bytes) -> bytes:
        return data

    def unprotect(self, data: bytes) -> bytes:
        return data

    def unprotect_rtcp(self, data: bytes) -> bytes:
        return data


class FakeDtlsTransport(RTCDtlsTransport):
    fake = True

    def getLocalParameters(self) -> RTCDtlsParameters:
        return RTCDtlsParameters(fingerprints=[])

    async def start(self, parameters: RTCDtlsParameters) -> None:
        self._ssl = FakeSsl()  # type: ignore
        self._rx_srtp = FakeDtlsSession()
        self._tx_srtp = FakeDtlsSession()

        self._set_state(State.CONNECTED)
        asyncio.ensure_future(self.__run())

    async def stop(self) -> None:
        pass

    async def __run(self) -> None:
        try:
            while True:
                await self._recv_next()
        except ConnectionError:
            for receiver in self._rtp_router.receivers:
                receiver._handle_disconnect()  # pragma: no cover
        finally:
            self._set_state(State.CLOSED)


def create_rtp_transceiver(transport: RTCDtlsTransport) -> RTCRtpTransceiver:
    kind = "audio"

    transceiver = RTCRtpTransceiver(
        direction="sendrecv",
        kind=kind,
        sender=RTCRtpSender(AudioStreamTrack(), transport),
        receiver=RTCRtpReceiver(kind, transport),
    )
    transceiver.receiver._set_rtcp_ssrc(transceiver.sender._ssrc)
    transceiver.receiver._track = RemoteStreamTrack(kind=kind)
    transceiver._codecs = [
        RTCRtpCodecParameters(
            mimeType="audio/PCMA", clockRate=8000, channels=1, payloadType=8
        )
    ]
    transceiver._transport = transport
    return transceiver


class Call(Dialog):
    """
    A SIP call with support for RTP.
    """

    def __init__(
        self,
        *,
        call_id: str,
        dialog_layer: DialogLayer,
        dtls: bool = False,
        local_address: sipmessage.Address,
        remote_address: sipmessage.Address,
        route_set: list[sipmessage.Address] = [],
        stun_server: str | None = None,
    ) -> None:
        super().__init__(
            call_id=call_id,
            dialog_layer=dialog_layer,
            local_address=local_address,
            remote_address=remote_address,
            route_set=route_set,
        )

        self.__remoteDtls: RTCDtlsParameters | None = None
        self.__remoteIce: RTCIceParameters | None = None
        self.__skip_ice = False
        self.__terminated = False

        # Create ICE transport.
        iceServers = []
        if stun_server:
            iceServers.append(RTCIceServer(stun_server))
        iceGatherer = RTCIceGatherer(iceServers=iceServers)
        self.__iceTransport = RTCIceTransport(iceGatherer)

        # Create DTLS transport.
        cls = RTCDtlsTransport if dtls else FakeDtlsTransport
        self.__dtlsTransport = cls(
            self.__iceTransport, [RTCCertificate.generateCertificate()]
        )

        # Create RTP sender / receiver.
        self.__rtpTransceiver = create_rtp_transceiver(self.__dtlsTransport)

    @property
    def audioTransceiver(self) -> RTCRtpTransceiver:
        """
        The RTP transceiver used for audio.
        """
        return self.__rtpTransceiver

    async def accept(self, request: sipmessage.Request) -> None:
        """
        Accept an incoming call by sending a `200` response.
        """
        self.__iceTransport._connection.ice_controlling = False

        await self._handle_sdp(request.body.decode())

        response = self.create_response(request, 200)
        response.content_type = "application/sdp"
        response.body = (await self._create_sdp()).encode()

        # Send response and start media flow.
        await self.send_response(response)
        await self._media_connect()

    async def hangup(self) -> None:
        """
        Hangup the call by sending a `BYE` request.
        """
        if self.__terminated:
            return

        request = self.create_request("BYE")
        await self.send_request(request)

        await self._media_close()

    async def invite(self, extra_headers: list[tuple[str, str]] = []) -> None:
        """
        Make an outgoing call by sending an `INVITE` request.

        If you wish to send additional headers in the `INVITE`, specify
        them in ``extra_headers``.
        """
        self.__iceTransport._connection.ice_controlling = True

        request = self.create_request("INVITE")
        request.content_type = "application/sdp"
        request.body = (await self._create_sdp()).encode()
        for key, value in extra_headers:
            request.headers.add(key, value)

        response = await self.send_request(request)
        if response.code >= 200 and response.code < 300:
            # Parse SDP and start media flow.
            await self._handle_sdp(response.body.decode())
            await self._media_connect()
        else:
            # Shutdown media and raise an exception.
            await self._media_close()
            raise Exception(f"Call failed: {response.code} {response.phrase}")

    async def handle_request(self, request: sipmessage.Request) -> None:
        if request.method == "BYE":
            response = self.create_response(request, 200)
            response.contact = [create_contact()]
            await self._dialog_layer._transaction_layer.send_response(response)
            await self._media_close()

    async def _create_sdp(self) -> str:
        await self.__iceTransport.iceGatherer.gather()

        ntp_seconds = clock.current_ntp_time() >> 32
        session = sdp.SessionDescription()
        session.origin = "- %d %d IN IP4 0.0.0.0" % (ntp_seconds, ntp_seconds)

        media = sdp.MediaDescription(
            kind=self.__rtpTransceiver.kind,
            port=9,
            profile=(
                "RTP/AVP"
                if hasattr(self.__dtlsTransport, "fake")
                else "UDP/TLS/RTP/SAVPF"
            ),
            fmt=[c.payloadType for c in self.__rtpTransceiver._codecs],
        )
        media.host = "0.0.0.0"
        media.direction = "sendrecv"

        # ice
        iceGatherer = self.__iceTransport.iceGatherer
        media.ice_candidates = iceGatherer.getLocalCandidates()
        media.ice_candidates_complete = iceGatherer.state == "completed"
        media.ice = iceGatherer.getLocalParameters()
        if media.ice_candidates:
            media.host = media.ice_candidates[0].ip
            media.port = media.ice_candidates[0].port

        # dtls
        media.dtls = self.__dtlsTransport.getLocalParameters()

        # rtp
        media.rtcp_host = "0.0.0.0"
        media.rtcp_port = 9
        media.rtcp_mux = True
        media.rtp.codecs = self.__rtpTransceiver._codecs

        session.media.append(media)

        return str(session)

    async def _handle_sdp(self, value: str) -> None:
        session = sdp.SessionDescription.parse(value)
        media = session.media[0]
        assert media.kind == "audio"

        # ice
        self.__remoteIce = media.ice
        for candidate in media.ice_candidates:
            await self.__iceTransport.addRemoteCandidate(candidate)
        if not media.ice_candidates:
            # remote host does not support ICE
            default_host = media.host or session.host
            default_port = media.port
            await self.__iceTransport.addRemoteCandidate(
                RTCIceCandidate(
                    foundation="default",
                    component=1,
                    ip=default_host,
                    protocol="udp",
                    priority=1,
                    port=default_port,
                    type="host",
                )
            )
            await self.__iceTransport.addRemoteCandidate(
                RTCIceCandidate(
                    foundation="default",
                    component=2,
                    ip=default_host,
                    protocol="udp",
                    priority=1,
                    port=default_port + 1,
                    type="host",
                )
            )
            self.__skip_ice = True

        # dtls
        self.__remoteDtls = media.dtls

    async def _media_connect(self) -> None:
        if self.__skip_ice:
            connection = self.__iceTransport._connection
            default_candidate = connection.get_default_candidate(COMPONENT_RTP)
            assert default_candidate is not None, "No default RTP candidate"
            connection.set_selected_pair(
                COMPONENT_RTP, default_candidate.foundation, "default"
            )
        else:
            await self.__iceTransport.start(self.__remoteIce)  # type: ignore
        await self.__dtlsTransport.start(self.__remoteDtls)  # type: ignore
        await self.__rtpTransceiver.sender.send(
            RTCRtpSendParameters(codecs=self.__rtpTransceiver._codecs)
        )
        await self.__rtpTransceiver.receiver.receive(
            RTCRtpReceiveParameters(codecs=self.__rtpTransceiver._codecs)
        )

    async def _media_close(self) -> None:
        await self.__rtpTransceiver.stop()
        await self.__dtlsTransport.stop()
        await self.__iceTransport.stop()
        self.__terminated = True
